/************************************************************************
   STEMMER IMAGING GmbH
 ------------------------------------------------------------------------
  Class    :  C# Toolbox3D.FileIO.AsciiPointList
  Autor    :  MaS
  Date     :  01/2013
  Purpose  :  Supply conversions between a X,Y,Z column based plaintext
              format and the Common Vision Blox Portable COP (Cloud Of 
              Points) format. The file extension is .asc
              
              The format of the text file complies to GFM .txt structure.
              
  Uses     :  Image DLL, Utilities DLL
  Hints    :  N/A
  Revision :  1.3.3
  Updates  :  N/A
  Extend   :  - Support more formats.
  Notes    :  This class is an exemplary implementation and will be 
              subject to future changes without notification. The interface 
              might change as well. It is intended to be integrated and
              modified in a users application. 
              It is supplied without warranty and not officially supported 
              by STEMMER IMAGING.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Globalization;

namespace Toolbox3D.FileIO
{
  class AsciiPointList : IDisposable
  {

#region Ctor/Dtor

    // Constructor
    public AsciiPointList()
    {

    }

    // Destructor
    ~AsciiPointList()
    {
      Dispose(false);
    }

    public void Close()
    {
      ((IDisposable)this).Dispose();
    }

    void IDisposable.Dispose()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }

    protected virtual void Dispose(bool fDisposing)
    {
        if (fDisposing)
        {
            // Free managed resources
        }
        // Free unmanaged resources
    }
         

#endregion


#region Public member functions (static)

    /// <summary>
    /// Deserialize a Portable COP from a list of points.
    /// </summary>
    /// <param name="inputFilePath">Name and path of the file to read from</param>
    /// <param name="outImg">Portable COP image to write to</param>
    public static void Read(String inputFilePath, out Cvb.Image.IMG pCOP)
    {
      Dictionary<String, String> headerLines;

      Read(inputFilePath, out pCOP, out headerLines);
    }


    /// <summary>
    /// Deserialize a Portable COP from a list of points.
    /// </summary>
    /// <param name="inputFilePath">Name and path of the file to read from</param>
    /// <param name="outImg">Portable COP image to write to</param>
    /// <param name="headerLines">File header info as key,value pairs</param>
    public static void Read(String inputFilePath, out Cvb.Image.IMG pCOP, out Dictionary<String, String> headerLines)
    {
      // Parse header to inquire image size
      Int32 sizeX = 0;
      Int32 sizeY = 0;

      headerLines = new Dictionary<String, String>();

      // Allocate lists
      List<float> X = new List<float>();
      List<float> Y = new List<float>();
      List<float> Z = new List<float>();

      try
      {
        // Parse data values from file
        parseAsciiPointListFile(inputFilePath, X, Y, Z, headerLines);

        // Get size parameters to check against
        String dictValue;
        headerLines.TryGetValue("Image dimension", out dictValue);
        char[] crossChar = "x".ToCharArray();
        string[] split = dictValue.Split(crossChar);
        
        Int32.TryParse(split[0], out sizeX);
        Int32.TryParse(split[1], out sizeY);
        
        // Check if number of entries equals number of image elements
        if (X.Count == sizeX * sizeY)
        {
          // Now try to create the Portable COP file
          createPortableCOP(sizeX, sizeY, X, Y, Z, out pCOP);
        }
        else
        {
          pCOP = IntPtr.Zero;
          throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.UnspecificError);
        }
      }
      catch (Exception ex)
      {
        throw new System.Runtime.Serialization.SerializationException("Deserialization failed: " + ex.Message.ToString());
      }
    }


    /// <summary>
    /// Serialize the a Portable COP dataset as a list of points.
    /// </summary>
    /// <param name="pCOP">Portable COP input data</param>
    /// <param name="outputFilePath">Name and path of the destination file</param>
    public static void Write(ref Cvb.Image.IMG pCOP, String outputFilePath)
    {
      // No empty header allowed
      // Dictionary<String,String> headerLines = null;
      
      Dictionary<String, String> headerLines = new Dictionary<String, String>();

      headerLines.Add("Image dimension", Cvb.Image.ImageWidth(pCOP).ToString() + "x" + Cvb.Image.ImageHeight(pCOP).ToString());

      Write(ref pCOP, headerLines, outputFilePath);
    }


    /// <summary>
    /// Serialize the a Portable COP dataset as a list of points.
    /// </summary>
    /// <param name="pCOP">Portable COP input data</param>
    /// <param name="headerLines">File header info as key,value pairs</param>
    /// <param name="outputFilePath">Name and path of the destination file</param>
    public static void Write(ref Cvb.Image.IMG pCOP, Dictionary<String, String> headerLines, String outputFilePath)
    {
      // Allocate lists
      List<float> X = new List<float>();
      List<float> Y = new List<float>();
      List<float> Z = new List<float>();

      try
      {
        // Parse Portable COP and write values into lists
        parsePortableCOP(ref pCOP, X, Y, Z);
        // Write values into output file
        createAsciiPointListFile(X, Y, Z, headerLines, outputFilePath);
      }
      catch (Exception ex)
      {
        throw new System.Runtime.Serialization.SerializationException("Serialization failed: " + ex.Message.ToString());
      }
    }


#endregion


#region Private member functions

    /// <summary>
    /// This function crafts a Portable COP image from 3 lists with X, Y and Z values
    /// </summary>
    /// <param name="sizeX">This is only temporary, future versions will parse the value from the file itself</param>
    /// <param name="sizeY">This is only temporary, future versions will parse the value from the file itself</param>
    /// <param name="X">List holding all X values</param>
    /// <param name="Y">List holding all Y values</param>
    /// <param name="Z">List holding all Z values</param>
    /// <param name="pCOP">The Portable COP output image</param>
    private static void createPortableCOP(Int32 sizeX, Int32 sizeY, 
        List<float> X, List<float> Y, List<float> Z, out Cvb.Image.IMG pCOP)
    {
        IntPtr basePtrR, basePtrG, basePtrB;
        Int32 xIncR, yIncR, xIncG, yIncG, xIncB, yIncB;

        // The target format is of datatype 800
        // 32 bit (Float):  32
        // Signed:         256
        // Floating point: 512
        Cvb.Image.CreateGenericImageDT(3, sizeX, sizeY, 800, out pCOP);

        // Get linear access to the base image
        if (Cvb.Utilities.GetLinearAccess(pCOP, 0, out basePtrR, out xIncR, out yIncR) &&
            Cvb.Utilities.GetLinearAccess(pCOP, 1, out basePtrG, out xIncG, out yIncG) &&
            Cvb.Utilities.GetLinearAccess(pCOP, 2, out basePtrB, out xIncB, out yIncB))
        {
          // Do preprocessing if necessary
          // X.Reverse();
          // Z.Reverse();
          // Y.Reverse();

          // Start processing
          unsafe
          {
            // Variables for the 3 color layers
            float* pixelR;
            float* pixelG;
            float* pixelB;

            for (int y = 0; y < sizeY; y++)
            {
              for (int x = 0; x < sizeX; x++)
              {
                // Get memory address
                pixelR = (float*)((int)basePtrR + x * xIncR + y * yIncR);
                pixelG = (float*)((int)basePtrG + x * xIncG + y * yIncG);
                pixelB = (float*)((int)basePtrB + x * xIncB + y * yIncB);

                // Write image data ! X and Y changed !
                *((float*)pixelR) = (float)(Math.Floor(Y[(y * sizeX) + x]));
                *((float*)pixelG) = (float)(Math.Floor(X[(y * sizeX) + x]));
                *((float*)pixelB) = (float)(Math.Floor(Z[(y * sizeX) + x]));
                
                /*
                // Write image data ! X and Y changed !
                *((float*)pixelR) = (float)(Math.Floor(Y[(y * sizeX) + x] * 10));
                *((float*)pixelG) = (float)(Math.Floor(X[(y * sizeX) + x] * 10));
                *((float*)pixelB) = (float)(Math.Floor(Z[(y * sizeX) + x] * 10));
                */
              }
            }
          }
        }
        else
        {
          throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.NotALinearImage);
        }
    }

    /// <summary>
    /// This function parses a text file containing 3 columns of floating point
    /// numbers in the order X, Y, Z.
    /// </summary>
    /// <param name="inputFilePath">Path and name of input point list to read from</param>
    /// <param name="X">List storing all X values</param>
    /// <param name="Y">List storing all Y values</param>
    /// <param name="Z">List storing all Z values</param>
    private static void parseAsciiPointListFile(String inputFilePath, List<float> X, List<float> Y, List<float> Z, Dictionary<String, String> headerLines)
    {
      StreamReader reader = null;

      // Select "." as decimal separator
      NumberFormatInfo nfi = new NumberFormatInfo();
      nfi.NumberGroupSeparator = "";
      nfi.NumberDecimalSeparator = ".";

      reader = new StreamReader(inputFilePath);
      String curLine;
      double x, y, z;

      char[] colonChar = ":".ToCharArray();
      char[] tabChar = "\t".ToCharArray();

      // Read the first header line
      Int32 numHeaderLines = 0;
      curLine = reader.ReadLine();
      string[] split = curLine.Split(colonChar);
      Int32.TryParse(split[1], out numHeaderLines);
      // If no header line is found, throw an exception
      if (numHeaderLines <= 0)
      {
        throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.NoArgument);
      }

      while ((curLine = reader.ReadLine()) != null)
      {
        // Only loop over the next block while there are header lines
        if (numHeaderLines > 0)
        {
          split = curLine.Split(colonChar);
          String dictParam = split[0];
          String dictValue = split[1];

          headerLines.Add(dictParam, dictValue);

          numHeaderLines--;
          continue;
        }

        // Change splitting character
        split = curLine.Split(tabChar);

        // Will only be true for non-header lines
        if (Double.TryParse(split[0], System.Globalization.NumberStyles.Float, nfi, out x) &&
            Double.TryParse(split[1], System.Globalization.NumberStyles.Float, nfi, out y) &&
            Double.TryParse(split[2], System.Globalization.NumberStyles.Float, nfi, out z))
        {
          // Add new coordinates to lists
          X.Add((float)x);
          Y.Add((float)y);
          Z.Add((float)z);
        }
      }

      reader.Dispose();
    }

    /// <summary>
    /// This function parses Portable COP data (3 planes, floating point, sorted X, Y, Z) and
    /// writes the parsed values into 3 dynamic lists.
    /// </summary>
    /// <param name="pCOP">Portable COP data to parse</param>
    /// <param name="X">List to write the X values to</param>
    /// <param name="Y">List to write the Y values to</param>
    /// <param name="Z">List to write the Z values to</param>
    private static void parsePortableCOP(ref Cvb.Image.IMG pCOP, List<float> X, List<float> Y, List<float> Z)
    {
      IntPtr basePtrR, basePtrG, basePtrB;
      Int32 xIncR, yIncR, xIncG, yIncG, xIncB, yIncB;

      // The source format is of datatype 800
      // 32 bit (Float):  32
      // Signed:         256
      // Floating point: 512
      if (!(Cvb.Image.ImageDatatype(pCOP, 0) == 800 && Cvb.Image.ImageDatatype(pCOP, 1) == 800 &&
        Cvb.Image.ImageDatatype(pCOP, 2) == 800))
        // Throw invalid datatype exception
        throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.InvalidDatatype);

                   
      // Get linear access to the base image
      if (Cvb.Utilities.GetLinearAccess(pCOP, 0, out basePtrR, out xIncR, out yIncR) &&
          Cvb.Utilities.GetLinearAccess(pCOP, 1, out basePtrG, out xIncG, out yIncG) &&
          Cvb.Utilities.GetLinearAccess(pCOP, 2, out basePtrB, out xIncB, out yIncB))
      {
        // Do preprocessing if necessary
        // X.Reverse();
        // Z.Reverse();
        // Y.Reverse();

        int sizeX = Cvb.Image.ImageWidth(pCOP);
        int sizeY = Cvb.Image.ImageHeight(pCOP);

        // Start processing
        unsafe
        {
          // Variables for the 3 color layers
          float* pixelR;
          float* pixelG;
          float* pixelB;

          for (int y = 0; y < sizeY; y++)
          {
            for (int x = 0; x < sizeX; x++)
            {
              // Get memory address
              pixelR = (float*)((int)basePtrR + x * xIncR + y * yIncR);
              pixelG = (float*)((int)basePtrG + x * xIncG + y * yIncG);
              pixelB = (float*)((int)basePtrB + x * xIncB + y * yIncB);

              // Write image data
              Y.Add(*pixelR);
              X.Add(*pixelG);
              Z.Add(*pixelB);

            }
          }
        } // end unsafe   
      }
      else
      {
        throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.NotALinearImage);
      }

    }


    /// <summary>
    /// This function creates a 3 column text file from lists provided to it. The file contains
    /// the X, Y and Z values linearly as float numbers.
    /// </summary>
    /// <param name="X">List holding all X values</param>
    /// <param name="Y">List holding all Y values</param>
    /// <param name="Z">List holding all Z values</param>
    /// <param name="outputFilePath">Name and path of the serialization file</param>
    private static void createAsciiPointListFile(List<float> X, List<float> Y, List<float> Z, Dictionary<String, String> headerLines, String outputFilePath)
    {
      Int32 listSize = X.Count;
      String dataPoint = null;
      System.IO.FileStream fs = null;

      // Select "." as decimal separator
      NumberFormatInfo nfi = new NumberFormatInfo();
      nfi.NumberGroupSeparator = "";
      nfi.NumberDecimalSeparator = ".";

      // Create file
      fs = System.IO.File.Create(outputFilePath);

      // Single header line
      String headerLine;
      byte[] byteArray;

      // Prepare conversion
      System.Text.ASCIIEncoding enc = new System.Text.ASCIIEncoding();

      if (headerLines == null)
      {
        // Empty header
        headerLine = "Header length: 0\r\n";

        // Convert data and write to file
        byteArray = enc.GetBytes(headerLine);
        fs.Write(byteArray, 0, headerLine.Length);
      }
      else
      {
        // Normal header
        headerLine = "Header length: " + headerLines.Count + "\r\n";

        // Prepare byte[] conversion, convert data and write to file
        byteArray = enc.GetBytes(headerLine);
        fs.Write(byteArray, 0, headerLine.Length);

        foreach (KeyValuePair<String, String> headerPair in headerLines)
        {
          headerLine = headerPair.Key + ": " + headerPair.Value + "\r\n";

          // Convert data and write to file
          byteArray = enc.GetBytes(headerLine);
          fs.Write(byteArray, 0, headerLine.Length);
        }
      }

      // Now write the image data to the file
      for (int i = 0; i < listSize; i++)
      {
        // Compile one line
        dataPoint = X[i].ToString("N6", nfi) + "\t" + Y[i].ToString("N6", nfi) + "\t" +
            Z[i].ToString("N6", nfi) + "\r\n";

        // Convert string into byte array
        byteArray = enc.GetBytes(dataPoint);

        // Write array into stream
        fs.Write(byteArray, 0, byteArray.Length);
      }
      fs.Close();
    }

#endregion

    }
}