﻿/************************************************************************
   STEMMER IMAGING GmbH
 ------------------------------------------------------------------------
  Class    :  C# Cvb.Exception
  Autor    :  MaS
  Date     :  12/2010
  Purpose  :  Provide CVB errors as exceptions
              
  Uses     :  N/A
  Hints    :  N/A
  Revision :  1.0
  Updates  :  N/A
  Extend   :  N/A
  Notes    :  This class is an exemplary implementation and will be 
              subject to future changes without notification. The interface 
              might change as well. It is intended to be integrated and
              modified in a users application. 
              It is supplied without warranty and not officially supported 
              by STEMMER IMAGING.
 ************************************************************************/

using System;
using System.Runtime.InteropServices;

namespace Cvb
{
  /// <summary>
  /// Predefined error values.
  /// </summary>
  /// <remarks>
  /// <para>The <see cref="System.Runtime.InteropServices.ExternalException.ErrorCode"/> property will have
  /// the most significant bit set to signal an error.</para>
  /// <para>Use the following C# code to get the enumeration value (if valid):
  /// <code>
  /// using Cvb.Toolbox3D;
  /// 
  /// try
  /// {
  ///   // Cvb operation
  /// }
  /// catch(Exception ex)
  /// {
  ///   CvbErrorCode code = (CvbErrorCode)(ex.ErrorCode &amp;0xFFFF);
  ///   MessageBox.Show(code.ToString());
  /// }
  /// </code>
  /// </para>
  /// </remarks>

  public enum CvbErrorCode
  {
    /// <summary>
    /// No error
    /// </summary>
    //Ok = 0,
    /// <summary>
    /// Unspecified error
    /// </summary>
    UnspecificError = 1,
    /// <summary>
    /// Invalid parameter
    /// </summary>
    InvalidParameter = 2,
    /// <summary>
    /// File I/O error
    /// </summary>
    FileIOError = 3,
    /// <summary>
    /// Acquisition timeout
    /// </summary>
    AcquisitionTimeout = 4,
    /// <summary>
    /// Memory allocation error
    /// </summary>
    MemoryError = 5,
    /// <summary>
    /// Invalid image plane
    /// </summary>
    InvalidPlane = 6,
    /// <summary>
    /// Unsupported datatype
    /// </summary>
    UnsupportedDatatype = 7,
    /// <summary>
    /// Invalid camera port
    /// </summary>
    InvalidCameraPort = 8,
    /// <summary>
    /// Board select failed
    /// </summary>
    BoardSelectFailed = 9,
    /// <summary>
    /// Invalid triggermode
    /// </summary>
    InvalidTriggerMode = 10,
    /// <summary>
    /// Could not read property
    /// </summary>
    PropertyReadError = 11,
    /// <summary>
    /// Could not write property
    /// </summary>
    PropertyWriteError = 12,
    /// <summary>
    /// Invalid port
    /// </summary>
    InvalidPort = 13,
    /// <summary>
    /// Could not read port
    /// </summary>
    PortReadError = 14,
    /// <summary>
    /// Could not write port
    /// </summary>
    PortWriteError = 15,
    /// <summary>
    /// Image not available
    /// </summary>
    NoImageAvailable = 16,
    /// <summary>
    /// Interface not available
    /// </summary>
    NoInterfaceAvailable = 17,
    /// <summary>
    /// Hardware busy
    /// </summary>
    HardwareBusy = 18,
    /// <summary>
    /// Feature is not supported
    /// </summary>
    FeatureNotSupported = 19,
    /// <summary>
    /// Grab has been aborted
    /// </summary>
    GrabAborted = 20,
    /// <summary>
    /// Pointer does not reference a pixel list
    /// </summary>
    NotAPixellist = 21,
    /// <summary>
    /// Not enough data available for computation
    /// </summary>
    NotEnoughData = 22,
    /// <summary>
    /// Object is not a NL-transformation object
    /// </summary>
    NotATransformation = 23,
    /// <summary>
    /// Function only works on linear images
    /// </summary>
    NotALinearImage = 24,
    /// <summary>
    /// Division by zero was attempted
    /// </summary>
    DivisionByZero = 25,
    /// <summary>
    /// Invalid number of planes
    /// </summary>
    InvalidDimension = 26,
    /// <summary>
    /// Invalid color model
    /// </summary>
    InvalidColorModel = 27,
    /// <summary>
    /// Invalid datatype
    /// </summary>
    InvalidDatatype = 28,
    /// <summary>
    /// Wrong object
    /// </summary>
    WrongObject = 29,
    /// <summary>
    /// Not ready for the requested operation
    /// </summary>
    NotReady = 30,
    /// <summary>
    /// Not an angle object
    /// </summary>
    NotAnAngle = 31,
    /// <summary>
    /// Not a 2D vector object
    /// </summary>
    NotA2DVector = 32,
    /// <summary>
    /// Not a 2D line object
    /// </summary>
    NotA2DLine = 33,
    /// <summary>
    /// Vector has zero length
    /// </summary>
    ZeroLengthVector = 34,
    /// <summary>
    /// Vectors are identical
    /// </summary>
    IdenticalVectors = 35,
    /// <summary>
    /// Line is vertical
    /// </summary>
    LineIsVertical = 36,
    /// <summary>
    /// Line is horizontal
    /// </summary>
    LineIsHorizontal = 37,
    /// <summary>
    /// Argument cannot be calculated for x = y = 0
    /// </summary>
    NoArgument = 38,
    /// <summary>
    /// Line 2D object has not yet been defined
    /// </summary>
    Undefined2DObject = 39,
    /// <summary>
    /// Lines provided to the function do not intersect
    /// </summary>
    NoIntersection = 40,
    /// <summary>
    /// No clipping points available
    /// </summary>
    NoClipping = 41,
    /// <summary>
    /// Not enough lines to perform intersection
    /// </summary>
    NotEnoughLines = 42,
    /// <summary>
    /// Input value was too big or did lead to a too big result
    /// </summary>
    OverflowError = 43,
    /// <summary>
    /// Object is not a 2D circle
    /// </summary>
    NotA2DCircle = 44,
    /// <summary>
    /// Access error
    /// </summary>
    AccessError = 45,
    /// <summary>
    /// Not present
    /// </summary>
    NotPresent = 46,
    /// <summary>
    /// Unsupported feature
    /// </summary>
    FeatureNotSupported2 = 47,
    /// <summary>
    /// Index exceeded limits
    /// </summary>
    InvalidIndex = 48,
    /// <summary>
    /// Image does not have overlay bits
    /// </summary>
    NoOverlayBits = 49
  }


  /// <summary>
  /// Reports an error that occurred in one of the CVB library methods.
  /// </summary>
  public class Exception
    : ExternalException
  {
    /// <summary>
    /// Creates an empty exception.
    /// </summary>
    public Exception() : base() { }

    /// <summary>
    /// Creates an exception with the given error <paramref name="message"/>.
    /// </summary>
    /// <param name="message">Reason for error.</param>
    public Exception(string message)
      : base(message)
    {
    }

    /// <summary>
    /// Creates an exception with the given error <paramref name="message"/>
    /// and <paramref name="errorCode"/>.
    /// </summary>
    /// <param name="message">Reason for error.</param>
    /// <param name="errorCode">Native error code.</param>
    public Exception(string message, int errorCode)
      : base(message, errorCode)
    {
    }

    /// <summary>
    /// This code extracts the case statements above from the initial enums
    /// 
    /// cat enum | perl -e 'my $msg; while (<>) { my $code = "", $line = $_; chomp($line); if ($line =~ m/\s*\/\/\/\s([^<]+)$/) { $msg = $1;} elsif ($line =~ m/\s*([A-Za-z0-9]+)\s.*/) {$code = $1;} if ($code) {print "case CvbErrorCode.".$code.":\n  return new Exception(\"".$msg."\", code);\n";} }' > switch
    /// </summary>

    /// <summary>
    /// Helper function for specific 3D error codes.
    /// </summary>
    /// <param name="code">CVB result type.</param>
    /// <returns>Exception object with appropriate description.</returns>
    public static void FromCvbErrorCode(int code)
    {
      CvbErrorCode enumCode = (CvbErrorCode)((uint)code & 0xFFFF);
      FromCvbErrorCode(enumCode);
    }

    /// <summary>
    /// Helper function for specific 3D error codes.
    /// </summary>
    /// <param name="code">CVB result type.</param>
    /// <returns>Exception object with appropriate description.</returns>
    public static Exception FromCvbErrorCode(CvbErrorCode enumCode)
    {
      int code = (int)enumCode;

      switch (enumCode)
      {
        // case CvbErrorCode.Ok:
        // break;
        case CvbErrorCode.UnspecificError:
          return new Exception("Unspecified error", code);
        case CvbErrorCode.InvalidParameter:
          return new Exception("Invalid parameter", code);
        case CvbErrorCode.FileIOError:
          return new Exception("File I/O error", code);
        case CvbErrorCode.AcquisitionTimeout:
          return new Exception("Acquisition timeout", code);
        case CvbErrorCode.MemoryError:
          return new Exception("Memory allocation error", code);
        case CvbErrorCode.InvalidPlane:
          return new Exception("Invalid image plane", code);
        case CvbErrorCode.UnsupportedDatatype:
          return new Exception("Unsupported datatype", code);
        case CvbErrorCode.InvalidCameraPort:
          return new Exception("Invalid camera port", code);
        case CvbErrorCode.BoardSelectFailed:
          return new Exception("Board select failed", code);
        case CvbErrorCode.InvalidTriggerMode:
          return new Exception("Invalid triggermode", code);
        case CvbErrorCode.PropertyReadError:
          return new Exception("Could not read property", code);
        case CvbErrorCode.PropertyWriteError:
          return new Exception("Could not write property", code);
        case CvbErrorCode.InvalidPort:
          return new Exception("Invalid port", code);
        case CvbErrorCode.PortReadError:
          return new Exception("Could not read port", code);
        case CvbErrorCode.PortWriteError:
          return new Exception("Could not write port", code);
        case CvbErrorCode.NoImageAvailable:
          return new Exception("Image not available", code);
        case CvbErrorCode.NoInterfaceAvailable:
          return new Exception("Interface not available", code);
        case CvbErrorCode.HardwareBusy:
          return new Exception("Hardware busy", code);
        case CvbErrorCode.FeatureNotSupported:
          return new Exception("Feature is not supported", code);
        case CvbErrorCode.GrabAborted:
          return new Exception("Grab has been aborted", code);
        case CvbErrorCode.NotAPixellist:
          return new Exception("Pointer does not reference a pixel list", code);
        case CvbErrorCode.NotEnoughData:
          return new Exception("Not enough data available for computation", code);
        case CvbErrorCode.NotATransformation:
          return new Exception("Object is not a NL-transformation object", code);
        case CvbErrorCode.NotALinearImage:
          return new Exception("Function only works on linear images", code);
        case CvbErrorCode.DivisionByZero:
          return new Exception("Division by zero was attempted", code);
        case CvbErrorCode.InvalidDimension:
          return new Exception("Invalid number of planes", code);
        case CvbErrorCode.InvalidColorModel:
          return new Exception("Invalid color model", code);
        case CvbErrorCode.InvalidDatatype:
          return new Exception("Invalid datatype", code);
        case CvbErrorCode.WrongObject:
          return new Exception("Wrong object", code);
        case CvbErrorCode.NotReady:
          return new Exception("Not ready for the requested operation", code);
        case CvbErrorCode.NotAnAngle:
          return new Exception("Not an angle object", code);
        case CvbErrorCode.NotA2DVector:
          return new Exception("Not a 2D vector object", code);
        case CvbErrorCode.NotA2DLine:
          return new Exception("Not a 2D line object", code);
        case CvbErrorCode.ZeroLengthVector:
          return new Exception("Vector has zero length", code);
        case CvbErrorCode.IdenticalVectors:
          return new Exception("Vectors are identical", code);
        case CvbErrorCode.LineIsVertical:
          return new Exception("Line is vertical", code);
        case CvbErrorCode.LineIsHorizontal:
          return new Exception("Line is horizontal", code);
        case CvbErrorCode.NoArgument:
          return new Exception("Argument cannot be calculated for x = y = 0", code);
        case CvbErrorCode.Undefined2DObject:
          return new Exception("Line 2D object has not yet been defined", code);
        case CvbErrorCode.NoIntersection:
          return new Exception("Lines provided to the function do not intersect", code);
        case CvbErrorCode.NoClipping:
          return new Exception("No clipping points available", code);
        case CvbErrorCode.NotEnoughLines:
          return new Exception("Not enough lines to perform intersection", code);
        case CvbErrorCode.OverflowError:
          return new Exception("Input value was too big or did lead to a too big result", code);
        case CvbErrorCode.NotA2DCircle:
          return new Exception("Object is not a 2D circle", code);
        case CvbErrorCode.AccessError:
          return new Exception("Access error", code);
        case CvbErrorCode.NotPresent:
          return new Exception("Not present", code);
        case CvbErrorCode.FeatureNotSupported2:
          return new Exception("Unsupported feature", code);
        case CvbErrorCode.InvalidIndex:
          return new Exception("Index exceeded limits", code);
        case CvbErrorCode.NoOverlayBits:
          return new Exception("Image does not have overlay bits", code);
        default:
          return new Exception("Unknown error (see status code)", code);
      }

    }
  }
}
