/************************************************************************
   STEMMER IMAGING GmbH
 ------------------------------------------------------------------------
  Class    :  C# Toolbox3D.FileIO.PointCloudData
  Autor    :  MaS
  Date     :  01/2013
  Purpose  :  Supply conversions between the ASCII form of the X,Y,Z column
              based LibPCL data format and the Common Vision Blox Portable 
              COP (Cloud Of Points) format. The file extension is .pcd.         
              
              See also:
              http://pointclouds.org/documentation/tutorials/pcd_file_format.php#pcd-file-format 
              
  Uses     :  Image DLL, Utilities DLL
  Hints    :  N/A
  Revision :  1.0.1
  Updates  :  N/A
  Extend   :  - Support binary format.
  Notes    :  This class is an exemplary implementation and will be 
              subject to future changes without notification. The interface 
              might change as well. It is intended to be integrated and
              modified in a users application. 
              It is supplied without warranty and not officially supported 
              by STEMMER IMAGING.
 ************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Globalization;

namespace Toolbox3D.FileIO
{
  class PointCloudData : IDisposable
  {

#region Ctor/Dtor

    // Constructor
    public PointCloudData()
    {

    }

    // Destructor
    ~PointCloudData()
    {
      Dispose(false);
    }

    public void Close()
    {
      ((IDisposable)this).Dispose();
    }

    void IDisposable.Dispose()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }

    protected virtual void Dispose(bool fDisposing)
    {
        if (fDisposing)
        {
            // Free managed resources
        }
        // Free unmanaged resources
    }
         

#endregion


#region Public member functions (static)

    /// <summary>
    /// Deserialize a Portable COP from a list of points.
    /// </summary>
    /// <param name="inputFilePath">Name and path of the file to read from</param>
    /// <param name="outImg">Portable COP image to write to</param>
    public static void Read(String inputFilePath, out Cvb.Image.IMG pCOP)
    {
      Dictionary<String, String> headerLines;

      Read(inputFilePath, out pCOP, out headerLines);
    }


    /// <summary>
    /// Deserialize a Portable COP from a list of points.
    /// </summary>
    /// <param name="inputFilePath">Name and path of the file to read from</param>
    /// <param name="outImg">Portable COP image to write to</param>
    /// <param name="headerLines">File header info as key,value pairs</param>
    public static void Read(String inputFilePath, out Cvb.Image.IMG pCOP, out Dictionary<String, String> headerLines)
    {
      // Parse header to inquire image size
      Int32 sizeX = 0;
      Int32 sizeY = 0;

      headerLines = new Dictionary<String, String>();

      // Allocate lists
      List<float> X = new List<float>();
      List<float> Y = new List<float>();
      List<float> Z = new List<float>();

      try
      {
        // Parse data values from file
        parsePointCloudDataFile(inputFilePath, headerLines);

        // Get size parameters to check against
        String width, height, fields, data;
        headerLines.TryGetValue("WIDTH", out width);
        headerLines.TryGetValue("HEIGHT", out height);
        headerLines.TryGetValue("FIELDS", out fields);
        headerLines.TryGetValue("DATA", out data);

        // Return immediately if format is not ASCII
        if (data != "ascii")
        {
          pCOP = IntPtr.Zero;
          throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.UnspecificError);
        }

        // Inspect type and number of fields
        char[] spaceChar = " ".ToCharArray();
        string[] split = fields.Split(spaceChar);      

        // Accept only "x y z" for now
        if (! (split.Length == 3 && 
          split[0] == "x" && split[1] == "y" && split[2] == "z"))
        { 
          pCOP = IntPtr.Zero;
          throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.UnspecificError);
        }

        Int32.TryParse(width, out sizeX);
        Int32.TryParse(height, out sizeY);  

        // Parse data
        parsePointCloudDataFile(inputFilePath, X, Y, Z);

        // Check if number of entries equals number of image elements
        if (X.Count == sizeX * sizeY)
        {
          // Now try to create the Portable COP file
          createPortableCOP(sizeX, sizeY, X, Y, Z, out pCOP);
        }
        else
        {
          pCOP = IntPtr.Zero;
          throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.UnspecificError);
        }
      }
      catch (Exception ex)
      {
        throw new System.Runtime.Serialization.SerializationException("Deserialization failed: " + ex.Message.ToString());
      }
    }


    /// <summary>
    /// Serialize the Portable COP dataset as PCD file with a fixed header.
    /// </summary>
    /// <param name="pCOP">Portable COP input data</param>
    /// <param name="outputFilePath">Name and path of the destination file</param>
    public static void Write(ref Cvb.Image.IMG pCOP, String outputFilePath)
    {
      // No empty header allowed
      // Dictionary<String,String> headerLines = null;
      
      Dictionary<String, String> headerLines = new Dictionary<String, String>();

      headerLines.Add("#", ".PCD v.7 - Point Cloud Data file format");
      headerLines.Add("VERSION", ".7");
      headerLines.Add("FIELDS", "x y z");
      headerLines.Add("SIZE", "4 4 4");
      headerLines.Add("TYPE", "F F F");
      headerLines.Add("COUNT", "1 1 1");
      headerLines.Add("WIDTH", Cvb.Image.ImageWidth(pCOP).ToString());
      headerLines.Add("HEIGHT", Cvb.Image.ImageHeight(pCOP).ToString());
      headerLines.Add("VIEWPOINT", "0 0 0 1 0 0 0");
      headerLines.Add("POINTS", (Cvb.Image.ImageWidth(pCOP) * Cvb.Image.ImageHeight(pCOP)).ToString());
      headerLines.Add("DATA", "ascii");

      Write(ref pCOP, headerLines, outputFilePath);
    }


    /// <summary>
    /// Serialize the a Portable COP dataset as PCD file.
    /// </summary>
    /// <param name="pCOP">Portable COP input data</param>
    /// <param name="headerLines">File header info as key,value pairs</param>
    /// <param name="outputFilePath">Name and path of the destination file</param>
    public static void Write(ref Cvb.Image.IMG pCOP, Dictionary<String, String> headerLines, String outputFilePath)
    {
      // Allocate lists
      List<float> X = new List<float>();
      List<float> Y = new List<float>();
      List<float> Z = new List<float>();

      try
      {
        // Parse Portable COP and write values into lists
        parsePortableCOP(ref pCOP, X, Y, Z);
        // Write values into output file
        createPointCloudDataFile(X, Y, Z, headerLines, outputFilePath);
      }
      catch (Exception ex)
      {
        throw new System.Runtime.Serialization.SerializationException("Serialization failed: " + ex.Message.ToString());
      }
    }

    public static void Write(ref Cvb.Image.IMG heightMap, ref Cvb.Image.IMG intensityMap,
      ref Cvb.Image.IMG validityMap, Double scaleX, Double scaleY, Double scaleZ,
      String outputFilePath)
    {
      // Allocate lists
      List<float> X = new List<float>();
      List<float> Y = new List<float>();
      List<float> Z = new List<float>();
      List<Int32> RGB = new List<Int32>();
      Dictionary<String, String> headerLines = new Dictionary<String, String>();

      try
      {
        // Parse Portable COP and write values into lists
        parseImageStack(ref heightMap, ref intensityMap, ref validityMap, 
          scaleX, scaleY, scaleZ, X, Y, Z, RGB);

        headerLines.Add("#", ".PCD v.7 - Point Cloud Data file format");
        headerLines.Add("VERSION", ".7");
        headerLines.Add("FIELDS", "x y z rgb");
        headerLines.Add("SIZE", "4 4 4 4");
        headerLines.Add("TYPE", "F F F F");
        headerLines.Add("COUNT", "1 1 1 1");
        headerLines.Add("WIDTH", X.Count.ToString());
        headerLines.Add("HEIGHT", "1");
        headerLines.Add("VIEWPOINT", "0 0 0 1 0 0 0");
        headerLines.Add("POINTS", X.Count.ToString());
        headerLines.Add("DATA", "ascii");

        // Write values into output file
        createPointCloudDataFile(X, Y, Z, RGB, headerLines, outputFilePath);
      }
      catch (Exception ex)
      {
        throw new System.Runtime.Serialization.SerializationException("Serialization failed: " + ex.Message.ToString());
      }
    }


#endregion


#region Private member functions

    /// <summary>
    /// This function crafts a Portable COP image from 3 lists with X, Y and Z values
    /// </summary>
    /// <param name="sizeX">This is only temporary, future versions will parse the value from the file itself</param>
    /// <param name="sizeY">This is only temporary, future versions will parse the value from the file itself</param>
    /// <param name="X">List holding all X values</param>
    /// <param name="Y">List holding all Y values</param>
    /// <param name="Z">List holding all Z values</param>
    /// <param name="pCOP">The Portable COP output image</param>
    private static void createPortableCOP(Int32 sizeX, Int32 sizeY, 
        List<float> X, List<float> Y, List<float> Z, out Cvb.Image.IMG pCOP)
    {
        IntPtr basePtrR, basePtrG, basePtrB;
        Int32 xIncR, yIncR, xIncG, yIncG, xIncB, yIncB;

        // The target format is of datatype 800
        // 32 bit (Float):  32
        // Signed:         256
        // Floating point: 512
        Cvb.Image.CreateGenericImageDT(3, sizeX, sizeY, 800, out pCOP);

        // Get linear access to the base image
        if (Cvb.Utilities.GetLinearAccess(pCOP, 0, out basePtrR, out xIncR, out yIncR) &&
            Cvb.Utilities.GetLinearAccess(pCOP, 1, out basePtrG, out xIncG, out yIncG) &&
            Cvb.Utilities.GetLinearAccess(pCOP, 2, out basePtrB, out xIncB, out yIncB))
        {
          // Do preprocessing if necessary
          // X.Reverse();
          // Z.Reverse();
          // Y.Reverse();

          // Start processing
          unsafe
          {
            // Variables for the 3 color layers
            float* pixelR;
            float* pixelG;
            float* pixelB;

            for (int y = 0; y < sizeY; y++)
            {
              for (int x = 0; x < sizeX; x++)
              {
                // Get memory address
                pixelR = (float*)((int)basePtrR + x * xIncR + y * yIncR);
                pixelG = (float*)((int)basePtrG + x * xIncG + y * yIncG);
                pixelB = (float*)((int)basePtrB + x * xIncB + y * yIncB);

                // Write image data ! X and Y changed !
                *((float*)pixelR) = (float)(Math.Floor(Y[(y * sizeX) + x]));
                *((float*)pixelG) = (float)(Math.Floor(X[(y * sizeX) + x]));
                *((float*)pixelB) = (float)(Math.Floor(Z[(y * sizeX) + x]));
                
                /*
                // Write image data ! X and Y changed !
                *((float*)pixelR) = (float)(Math.Floor(Y[(y * sizeX) + x] * 10));
                *((float*)pixelG) = (float)(Math.Floor(X[(y * sizeX) + x] * 10));
                *((float*)pixelB) = (float)(Math.Floor(Z[(y * sizeX) + x] * 10));
                */
              }
            }
          }
        }
        else
        {
          throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.NotALinearImage);
        }
    }

    /// <summary>
    /// This function parses a PCD file header.
    /// </summary>
    /// <param name="inputFilePath">Path and name of input PCD file to read from</param>
    /// <param name="headerLines">List of header lines</param>
    private static void parsePointCloudDataFile(String inputFilePath, Dictionary<String, String> headerLines)
    {
      StreamReader reader = null;
      float canary = 0;

      reader = new StreamReader(inputFilePath);
      String curLine;

      string[] split = null;
      char[] spaceChar = " ".ToCharArray();

      String dictParam = "";
      String dictValue = "";

      do {
        curLine = reader.ReadLine();

        split = curLine.Split(spaceChar);

        dictParam = split[0];
        for (Int32 i = 1; i < split.Length; i++)
          dictValue += (split[i] + " ");

        headerLines.Add(dictParam, dictValue);
      } while (!(float.TryParse(dictParam, out canary)));
      
      // Check if anything has be read at all
      if (headerLines.Count <= 0)
      {
        throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.NoArgument);
      }

      reader.Dispose();
    }

    /// <summary>
    /// This function parses a text file containing 3 columns of floating point
    /// numbers in the order X, Y, Z.
    /// </summary>
    /// <param name="inputFilePath">Path and name of input point list to read from</param>
    /// <param name="X">List storing all X values</param>
    /// <param name="Y">List storing all Y values</param>
    /// <param name="Z">List storing all Z values</param>
    private static void parsePointCloudDataFile(String inputFilePath, List<float> X, List<float> Y, List<float> Z)
    {
      StreamReader reader = null;

      // Select "." as decimal separator
      NumberFormatInfo nfi = new NumberFormatInfo();
      nfi.NumberGroupSeparator = "";
      nfi.NumberDecimalSeparator = ".";

      reader = new StreamReader(inputFilePath);
      String curLine;
      string[] split;
      double x, y, z;

      char[] spaceChar = " ".ToCharArray();

      // Skip the header
      while ((curLine = reader.ReadLine()) != null)
      {
        // Change splitting character
        split = curLine.Split(spaceChar);

        // Will only be true for non-header lines
        if (Double.TryParse(split[0], System.Globalization.NumberStyles.Float, nfi, out x) &&
            Double.TryParse(split[1], System.Globalization.NumberStyles.Float, nfi, out y) &&
            Double.TryParse(split[2], System.Globalization.NumberStyles.Float, nfi, out z))
        {
          // Add new coordinates to lists
          X.Add((float)x);
          Y.Add((float)y);
          Z.Add((float)z);
        }
      }

      reader.Dispose();
    }

    /// <summary>
    /// This function parses Portable COP data (3 planes, floating point, sorted X, Y, Z) and
    /// writes the parsed values into 3 dynamic lists.
    /// </summary>
    /// <param name="pCOP">Portable COP data to parse</param>
    /// <param name="X">List to write the X values to</param>
    /// <param name="Y">List to write the Y values to</param>
    /// <param name="Z">List to write the Z values to</param>
    private static void parsePortableCOP(ref Cvb.Image.IMG pCOP, List<float> X, List<float> Y, List<float> Z)
    {
      IntPtr basePtrR, basePtrG, basePtrB;
      Int32 xIncR, yIncR, xIncG, yIncG, xIncB, yIncB;

      // The source format is of datatype 800
      // 32 bit (Float):  32
      // Signed:         256
      // Floating point: 512
      if (!(Cvb.Image.ImageDatatype(pCOP, 0) == 800 && Cvb.Image.ImageDatatype(pCOP, 1) == 800 &&
        Cvb.Image.ImageDatatype(pCOP, 2) == 800))
        // Throw invalid datatype exception
        throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.InvalidDatatype);

                   
      // Get linear access to the base image
      if (Cvb.Utilities.GetLinearAccess(pCOP, 0, out basePtrR, out xIncR, out yIncR) &&
          Cvb.Utilities.GetLinearAccess(pCOP, 1, out basePtrG, out xIncG, out yIncG) &&
          Cvb.Utilities.GetLinearAccess(pCOP, 2, out basePtrB, out xIncB, out yIncB))
      {
        // Do preprocessing if necessary
        // X.Reverse();
        // Z.Reverse();
        // Y.Reverse();

        int sizeX = Cvb.Image.ImageWidth(pCOP);
        int sizeY = Cvb.Image.ImageHeight(pCOP);

        // Start processing
        unsafe
        {
          // Variables for the 3 color layers
          float* pixelR;
          float* pixelG;
          float* pixelB;

          for (int y = 0; y < sizeY; y++)
          {
            for (int x = 0; x < sizeX; x++)
            {
              // Get memory address
              pixelR = (float*)((int)basePtrR + x * xIncR + y * yIncR);
              pixelG = (float*)((int)basePtrG + x * xIncG + y * yIncG);
              pixelB = (float*)((int)basePtrB + x * xIncB + y * yIncB);

              // Write image data
              Y.Add(*pixelR);
              X.Add(*pixelG);
              Z.Add(*pixelB);

            }
          }
        } // end unsafe   
      }
      else
      {
        throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.NotALinearImage);
      }

    }



    /// <summary>
    /// This function takes 3 congruent 16 bit unsigned images as well as scaling factors
    /// and extracts the respective points in 3D space.
    /// </summary>
    /// <param name="heightMap">16 bit unsigned image with 16 bit distance data</param>
    /// <param name="intensityMap">16 bit unsigned image with 8 bit intensity data</param>
    /// <param name="validityMap">16 bit unsigned image with binary data contents</param>
    /// <param name="X">Scaling factor for X coordinates</param>
    /// <param name="Y">Scaling factor for Y coordinates</param>
    /// <param name="Z">Scaling factor for Z coordinates</param>
    /// <param name="X">List holding all X values</param>
    /// <param name="Y">List holding all Y values</param>
    /// <param name="Z">List holding all Z values</param>
    /// <param name="RGB">List holding all RGB values</param>
    private static void parseImageStack(ref Cvb.Image.IMG depthMap, ref Cvb.Image.IMG intensityMap, 
      ref Cvb.Image.IMG validityMap, Double scaleX, Double scaleY, Double scaleZ,
      List<float> X, List<float> Y, List<float> Z, List<Int32> RGB)
    {
      
      IntPtr basePtrValid, basePtrDepth, basePtrIntensity;
      Int32 xIncValid, yIncValid, xIncDepth, yIncDepth, xIncIntensity, yIncIntensity;

      // Check if all images are 16 bit
      if (!(Cvb.Image.ImageDatatype(depthMap, 0) == 16 && Cvb.Image.ImageDatatype(validityMap, 0) == 16 &&
        Cvb.Image.ImageDatatype(intensityMap, 0) == 16))
        // Throw invalid datatype exception
        throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.InvalidDatatype);

      // Get linear access to the images 
      if (Cvb.Utilities.GetLinearAccess(validityMap, 0, out basePtrValid, out xIncValid, out yIncValid) &&
          Cvb.Utilities.GetLinearAccess(depthMap, 0, out basePtrDepth, out xIncDepth, out yIncDepth) &&
          Cvb.Utilities.GetLinearAccess(intensityMap, 0, out basePtrIntensity, out xIncIntensity, out yIncIntensity))
      {

        // Iterate over the validity map to find valid coordinates
        int sizeX = Cvb.Image.ImageWidth(validityMap);
        int sizeY = Cvb.Image.ImageHeight(validityMap);

        // Start processing
        unsafe
        {
          // Adress for validity data
          ushort* dataValid;
          ushort* dataDepth;
          ushort* dataIntensity;

          for (int y = 0; y < sizeY; y++)
          {
            for (int x = 0; x < sizeX; x++)
            {
              dataValid = (ushort*)((int)basePtrValid + x * xIncValid + y * yIncValid);

              // Insert data only if valid
              if (*dataValid == 1)
              {
                // Get memory address
                dataDepth = (ushort*)((int)basePtrDepth + x * xIncDepth + y * yIncDepth);
                dataIntensity = (ushort*)((int)basePtrIntensity + x * xIncIntensity + y * yIncIntensity);

                // Write image data
                X.Add((float)(x * scaleX));
                Y.Add((float)(y * scaleY));
                Z.Add((float)(*dataDepth * scaleZ));
                RGB.Add(*dataIntensity << 16 | *dataIntensity << 8 | *dataIntensity);
              }


            }
          }
        } // end unsafe   
      }
      else
      {
        throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.NotALinearImage);
      }
    }


    /// <summary>
    /// This function creates a 3 column text file from lists provided to it. The file contains
    /// the X, Y and Z values linearly as float numbers.
    /// </summary>
    /// <param name="X">List holding all X values</param>
    /// <param name="Y">List holding all Y values</param>
    /// <param name="Z">List holding all Z values</param>
    /// <param name="headerLines">Prefix header data</param>
    /// <param name="outputFilePath">Name and path of the serialization file</param>
    private static void createPointCloudDataFile(List<float> X, List<float> Y, List<float> Z, 
      Dictionary<String, String> headerLines, String outputFilePath)
    {
      // Create dummy RGB list
      // Intensity values are omitted if this list equals NULL
      List<Int32> RGB = null;

      createPointCloudDataFile(X, Y, Z, RGB, headerLines, outputFilePath);
    }


    /// <summary>
    /// This function creates a 3 column text file from lists provided to it. The file contains
    /// the X, Y and Z values linearly as float numbers as well as RGB in a packed format
    /// (see conversion functions at end of file).
    /// </summary>
    /// <param name="X">List holding all X values</param>
    /// <param name="Y">List holding all Y values</param>
    /// <param name="Z">List holding all Z values</param>
    /// <param name="RGB">List holding encoded RGB values</param>
    /// <param name="headerLines">Prefix header data</param>
    /// <param name="outputFilePath">Name and path of the serialization file</param>
    private static void createPointCloudDataFile(List<float> X, List<float> Y, List<float> Z,
      List<Int32> RGB, Dictionary<String, String> headerLines, String outputFilePath)
    {
      Int32 listSize = X.Count;
      String dataPoint = null;
      System.IO.FileStream fs = null;

      // Select "." as decimal separator
      NumberFormatInfo nfi = new NumberFormatInfo();
      nfi.NumberGroupSeparator = "";
      nfi.NumberDecimalSeparator = ".";

      // Create file
      fs = System.IO.File.Create(outputFilePath);

      // Single header line
      String headerLine;
      byte[] byteArray;

      // Prepare conversion
      System.Text.ASCIIEncoding enc = new System.Text.ASCIIEncoding();

      if (headerLines == null)
      {
        throw Cvb.Exception.FromCvbErrorCode(Cvb.CvbErrorCode.NoArgument);
      }
      else
      {
        // Prepare byte[] conversion, convert data and write to file
        foreach (KeyValuePair<String, String> headerPair in headerLines)
        {
          headerLine = headerPair.Key + ": " + headerPair.Value + "\n";

          // Convert data and write to file
          byteArray = enc.GetBytes(headerLine);
          fs.Write(byteArray, 0, headerLine.Length);
        }
      }

      // Depending on the type of data, write only positions or position + RGB
      // Put into two separate structures due to performance considerations
      if (RGB == null)
      {
        for (int i = 0; i < listSize; i++)
        {
            // Only write positions
            dataPoint = String.Format(nfi, "{0:N6} {1:N6} {2:N6}\n", X[i], Y[i], Z[i]);

            // Convert NaN to nan
            dataPoint = dataPoint.ToLower();

            // Convert string into byte array
            byteArray = enc.GetBytes(dataPoint);

            // Write array into stream
            fs.Write(byteArray, 0, byteArray.Length);
        }
      }
      else
      {
        // Now write the image data to the file
        for (int i = 0; i < listSize; i++)
        {
          // Write position and intensity
          dataPoint = String.Format(nfi, "{0:N6} {1:N6} {2:N6} {3:e}\n", X[i], Y[i], Z[i], RGB[i]);

          // Convert NaN to nan
          dataPoint = dataPoint.ToLower();

          // Convert string into byte array
          byteArray = enc.GetBytes(dataPoint);

          // Write array into stream
          fs.Write(byteArray, 0, byteArray.Length);
        }
      }

      fs.Close();
    }


    private static void unpackRGB(Int32 RGB, ref Int32 R, ref Int32 G, ref Int32 B)
    {
      R = (RGB >> 16) & 0x0000ff;
      G = (RGB >> 8) & 0x0000ff;
      B = (RGB) & 0x0000ff;
    }

    private static Int32 packRGB(Int32 R, Int32 G, Int32 B)
    {
      return (R << 16 | G << 8 | B);
    }

#endregion

    }
}